from datetime import date
from datetime import timedelta

from CleanEmonCore.Events import Observer
from CleanEmonCore.Events.builtins import DateChange

from ..lib.DBConnector import fetch_data
from ..lib.DBConnector import send_data

from ..Disaggregator import energy_data_to_dataframe
from ..Disaggregator import dataframe_to_energy_data
from ..Disaggregator import disaggregate


def update(yesterday: str):
    energy_data = fetch_data(yesterday)
    df = energy_data_to_dataframe(energy_data)

    df = disaggregate(df)
    dis_energy_data = dataframe_to_energy_data(df)
    send_data(yesterday, dis_energy_data)


def run():
    class Updater(Observer):
        def on_notify(self, *args, **kwargs):
            if "date" in kwargs:
                yesterday = kwargs["date"]
            else:  # By default, get the previous date
                yesterday = date.today() - timedelta(days=1)
            update(str(yesterday))

    event = DateChange(3, initial_date=date.today())  # todo: increase interval to reduce execution time?
    Updater(event)

    event.run()



from CleanEmonCore import CONFIG_FILE
from CleanEmonCore.CouchDBAdapter import CouchDBAdapter

from CleanEmonBackend.Disaggregator.service import update

adapter = CouchDBAdapter(CONFIG_FILE)
print(f"You are working on database: {adapter.db}")


def _disaggregate(date: str):
    print(f"Working on {date}")
    print("Disaggregating...")
    update(date)
    print("Done")


def disaggregate(*dates: str, no_prompt=False):
    for date in dates:
        if no_prompt:
            ans = True
        else:
            ans = input(f"Proceed with {date}? (<enter>: no) ")

        if ans:
            _disaggregate(date)
        else:
            break



"""This module defines the core functionality of the API"""

import os
from datetime import datetime
from datetime import timedelta

from typing import List
from typing import Dict

from CleanEmonCore.models import EnergyData

from .. import RES_DIR
from ..lib.DBConnector import fetch_data
from ..lib.DBConnector import adapter
from ..lib.plots import plot_data


def get_data(date: str, from_cache: bool, sensors: List[str] = None) -> EnergyData:
    """Fetches and prepares the daily data that will be returned, filtering in the provided `sensors`.
    Note that there is no need to explicitly specify the "timestamp sensor", as it will always be included.

    date -- a valid date string in `YYYY-MM-DD` format
    from_cache -- specifies whether the data should be searched in cache first. This may speed up the response time
    sensors -- an inclusive list containing the values of interest
    """

    raw_data = fetch_data(date, from_cache=from_cache).energy_data

    if sensors:
        if "timestamp" not in sensors:
            sensors.append("timestamp")

        filtered_data = []
        for record in raw_data:
            filtered_record = {sensor: value for sensor, value in record.items() if sensor in sensors}
            filtered_data.append(filtered_record)
        data = filtered_data
    else:
        data = raw_data

    return EnergyData(date, data)


def get_range_data(from_date: str, to_date: str, use_cache: bool, sensors: List[str] = None) -> Dict:
    """Fetches and prepares the range data that will be returned.

    from_date -- a valid date string in `YYYY-MM-DD` format
    to_date -- a valid date string in `YYYY-MM-DD` format. It MUST be chronologically greater or equal to `from_date`
    from_cache -- specifies whether the data should be searched in cache first. This may speed up the response time
    sensors -- an inclusive list containing the values of interest
    """

    # Define the range data schema
    # todo: maybe define a an appropriate solid dataclass?
    data = {
        "from_date": from_date,
        "to_date": to_date,
        "range_data": []
    }

    from_dt = datetime.strptime(from_date, "%Y-%m-%d")
    to_dt = datetime.strptime(to_date, "%Y-%m-%d")
    one_day = timedelta(days=1)

    # Concatenate energy data from multiple dates into a single list
    now = from_dt
    while now <= to_dt:
        now_str = now.strftime("%Y-%m-%d")
        daily_data = get_data(now_str, use_cache, sensors)
        data["range_data"].append(daily_data)
        now += one_day

    return data


def get_plot(date: str, from_cache: bool, sensors: List[str] = None) -> str:
    """Fetches and plots the desired data. Returns the path of the resulting plot.

    date -- a valid date string in `YYYY-MM-DD` format
    from_cache -- specifies whether the data should be searched in cache first. This may speed up the response time
    sensors -- an inclusive list containing the values of interest
    """

    energy_data = get_data(date, from_cache, sensors)
    f_out = plot_data(energy_data, columns=sensors)

    return os.path.join(RES_DIR, f_out)


def get_date_consumption(date: str, from_cache: bool, simplify: bool):
    """Hardcoded fetch-prepare accumulator function that handles the daily KwH. Returns the daily consumption in kwh.

    Acts as an under-the-curve measurement by subtracting the lowest power measurement from the highest one.
    It's not given that the first record of the energy data will always contain valid power values and thus, the "first
    value" is actually searched and cherry-picked. Same goes for the "last valid value".

    date -- a valid date string in `YYYY-MM-DD` format
    from_cache -- specifies whether the data should be searched in cache first. This may speed up the response time
    simplify -- if true, returns a single value, not a JSON object
    """

    data = get_data(date, from_cache)

    kwh_list = [record["kwh"] for record in data.energy_data]

    # Find the first valid kwh measurement
    first_valid = 0
    for kwh in kwh_list:
        if kwh:
            first_valid = kwh
            break

    # Find the last valid kwh measurement
    last_valid = 0
    for kwh in reversed(kwh_list):
        if kwh:
            last_valid = kwh
            break

    consumption = last_valid - first_valid
    if simplify:
        data = consumption
    else:
        data = {
            "consumption": consumption,
            "unit": "kwh"
        }

    return data


def get_meta(field: str = None) -> Dict:
    meta = adapter.fetch_meta()
    if not field:
        return meta
    else:
        if field in meta:
            return meta[field]
    return {}



from CleanEmonBackend.Disaggregator.preparation import energy_data_to_dataframe


def test_energy_data_to_dataframe(energy_data):
    old_cols = energy_data.energy_data[0].keys()
    df = energy_data_to_dataframe(energy_data)

    assert len(df.columns) == len(old_cols) + 1
    assert df.shape[0] == 17280
    assert any([("original" in col) for col in df.columns])



from CleanEmonBackend.lib.validation import is_valid_date
from CleanEmonBackend.lib.validation import is_valid_date_range


def test_is_valid_date():
    assert is_valid_date("2022-01-01")
    assert is_valid_date("2522-01-01")

    assert not is_valid_date("2022")
    assert not is_valid_date("2022-10")
    assert not is_valid_date("2022-13-01")
    assert not is_valid_date("")


def test_is_valid_date_range():
    assert is_valid_date_range("2022-01-01", "2022-01-01")
    assert is_valid_date_range("2022-01-01", "2022-01-02")
    assert is_valid_date_range("2022-01-01", "2022-12-01")
    assert is_valid_date_range("2022-01-01", "2025-01-01")

    assert not is_valid_date_range("2022-01-02", "2022-01-01")
    assert not is_valid_date_range("2022-01-01", "2022")
    assert not is_valid_date_range("2022", "2022-01-01")




"""This defines the FastAPI boostrap function"""

import datetime
from typing import Optional

from fastapi import FastAPI
from fastapi import Request
from fastapi.responses import JSONResponse
from fastapi.responses import FileResponse


def create_app():
    """Creates the FastAPI app"""

    from .API import get_data
    from .API import get_range_data
    from .API import get_date_consumption
    from .API import get_plot
    from .API import get_meta

    from ..lib.exceptions import BadDateError
    from ..lib.exceptions import BadDateRangeError

    from ..lib.validation import is_valid_date
    from ..lib.validation import is_valid_date_range

    meta_tags = [
        {
            "name": "Views",
            "description": "Essential views"
        },
        {
            "name": "Experimental",
            "description": "Cutting-edge features that may not be stable yet"
        }
    ]

    app = FastAPI(openapi_tags=meta_tags, swagger_ui_parameters={"defaultModelsExpandDepth": -1})

    def parse_date(date: str) -> str:
        """Simple date parser. A date can either be in a standard YYYY-MM-DD format or a predefined alias.
        If the given date is invalid, a BadDateError is being raised.
        """

        parsed_date: str

        if date.lower() == "today":
            parsed_date = datetime.date.today().isoformat()
        elif date.lower() == "yesterday":
            yesterday = datetime.date.today() - datetime.timedelta(days=1)
            parsed_date = yesterday.isoformat()
        elif is_valid_date(date):
            parsed_date = date
        else:
            raise BadDateError(date)

        return parsed_date

    @app.exception_handler(BadDateError)
    def bad_date_exception_handler(request: Request, exception: BadDateError):
        return JSONResponse(
            status_code=400,
            content={"message": f"Bad date ({exception.bad_date}), not in ISO format (YYYY-MM-DD)."}
        )

    @app.exception_handler(BadDateRangeError)
    def bad_date_range_exception_handler(request: Request, exception: BadDateRangeError):
        return JSONResponse(
            status_code=400,
            content={"message": f"Bad date range ({exception.bad_from_date} - {exception.bad_to_date}). Dates must "
                                f"be in ISO format (YYYY-MM-DD) and placed in correct order."}
        )

    @app.get("/json/date/{date}", tags=["Views"])
    def get_json_date(date: str = None, from_cache: bool = False, sensors: Optional[str] = None):
        """Returns the daily data for the supplied **{date}**.

        - **{date}**: A date in YYYY-MM-DD format
        - **from_cache**: If set to False, forces data to be fetched again from the central database. If set to True,
        data will be looked up in cache and then, if they are not found, fetched from the central database.
        - **sensors**: A comma (,) separated list of sensors to be returned. If present, only sensors defined in that
        list will be returned
        """

        parsed_date = parse_date(date)

        if sensors:
            sensors = sensors.split(',')

        return get_data(parsed_date, from_cache, sensors)

    @app.get("/json/range/{from_date}/{to_date}", tags=["Views"])
    def get_json_range(from_date: str, to_date: str, from_cache: bool = False,
                       sensors: Optional[str] = None):
        """Returns the range data for the supplied range, from **{from_date}** to **{to_date}**.

        - **{from_date}**: A date in YYYY-MM-DD format
        - **to_date**: A date in YYYY-MM-DD format. It should be chronologically greater or equal to **{from_date}**
        - **from_cache**: If set to False, forces data to be fetched again from the central database. If set to True,
        data will be looked up in cache and then, if they are not found, fetched from the central database.
        - **sensors**: A comma (,) separated list of sensors to be returned. If present, only sensors defined in that
        list will be returned
        """

        if not is_valid_date_range(from_date, to_date):
            raise BadDateRangeError(from_date, to_date)

        if sensors:
            sensors = sensors.split(',')

        return get_range_data(from_date, to_date, from_cache, sensors)

    @app.get("/plot/date/{date}", tags=["Experimental"])
    def get_plot_date(date: str = None, from_cache: bool = False, sensors: Optional[str] = None):
        """Returns the plot of the specified data, as a JPEG image.

        - **{date}**: A date in YYYY-MM-DD format
        - **from_cache**: If set to False, forces data to be fetched again from the central database. If set to True,
        data will be looked up in cache and then, if they are not found, fetched from the central database.
        - **sensors**: A comma (,) separated list of sensors to be returned. If present, only sensors defined in that
        list will be returned
        """

        parsed_date = parse_date(date)

        if sensors:
            sensors = sensors.split(',')

        plot_path = get_plot(parsed_date, from_cache, sensors)

        return FileResponse(plot_path, media_type="image/jpeg")

    @app.get("/plot/range/{from_date}/{to_date}", tags=["Experimental"])
    def get_plot_range(from_date: str, to_date: str, from_cache: bool = False,
                       sensors: Optional[str] = None):
        """Under construction :)"""
        return JSONResponse(
            status_code=501,
            content={"message": "This feature is currently not implemented"}
        )

    @app.get("/json/date/{date}/consumption", tags=["Views"])
    def get_json_date_consumption(date: str = None, from_cache: bool = False, simplify: bool = False):
        """Returns the power consumption for the given date.

        - **{date}**: A date in YYYY-MM-DD format
        - **from_cache**: If set to False, forces data to be fetched again from the central database. If set to True,
        data will be looked up in cache and then, if they are not found, fetched from the central database.
        """

        parsed_date = parse_date(date)

        return get_date_consumption(parsed_date, from_cache, simplify)

    @app.get("/meta/", tags=["Experimental"])
    @app.get("/meta/{field}", tags=["Experimental"])
    def get_json_meta(field: str = None):
        """Returns the metadata for the current house.

        - **{meta}**: Optional endpoint that specifies the field to be returned if it exists in metadata, otherwise an
        empty dict will be returned. If omitted, all meta fields will be returned.
        """

        return get_meta(field)

    return app



import pytest

from CleanEmonBackend.Disaggregator.inference import disaggregate


@pytest.mark.slow
def test_disaggregate(dataframe):
    df = disaggregate(dataframe)

    assert df.shape[0] == 17280
    assert len(df.columns) > len(dataframe.columns)



import pytest
from fastapi.testclient import TestClient

from CleanEmonBackend.API import create_app

api = create_app()
client = TestClient(api)


class TestGetData:

    def test_json_date(self):
        for date in ["today", "YESTERDAY", "2022-05-01"]:
            response = client.get(f"/json/date/{date}")
            data = response.json()

            assert response.status_code == 200

            assert "date" in data
            assert type(data["date"]) is str

            assert "energy_data" in data
            assert type(data["energy_data"]) is list

    def test_json_range(self):
        response = client.get("/json/range/2022-05-01/2022-05-02")
        data = response.json()

        assert response.status_code == 200

        assert "from_date" in data
        assert "to_date" in data
        assert "range_data" in data
        assert type(data["range_data"]) is list

    @pytest.mark.projectwise
    def test_sensors(self):
        sensors = "timestamp,power,external_temp"

        response = client.get(f"/json/date/2022-05-10?sensors={sensors}")
        data = response.json()

        for sensor in sensors.split(","):
            assert sensor in data["energy_data"][0]

    def test_bad_date(self):
        for today_str in ["last_year", "20-2-3", "2020-06-34"]:

            response = client.get(f"/json/date/{today_str}")
            data = response.json()

            assert response.status_code == 400

            assert "message" in data
            assert today_str in data["message"]

    def test_bad_date_range_1(self):
        """The to_date is an invalid date"""

        to_date = "2000"

        response = client.get(f"/json/range/2022-05-01/{to_date}")
        data = response.json()

        assert response.status_code == 400

        assert "message" in data
        assert to_date in data["message"]

    def test_bad_date_range_2(self):
        """The first date is valid but the second date is valid, but in bad order."""
        to_date = "2000-01-01"

        response = client.get(f"/json/range/2022-05-01/{to_date}")
        data = response.json()

        assert response.status_code == 400

        assert "message" in data
        assert to_date in data["message"]



import os
import json

import pandas as pd
import pytest
from CleanEmonCore.models import EnergyData
from CleanEmonBackend.Disaggregator.preparation import energy_data_to_dataframe


@pytest.fixture(scope="package")
def energy_data() -> EnergyData:
    directory = os.path.dirname(__file__)
    file = os.path.join(directory, "energy_data.json")

    with open(file, "r") as fp:
        data_dict = json.load(fp)

    return EnergyData(data_dict["date"], data_dict["energy_data"])


@pytest.fixture(scope="package")
def dataframe(energy_data) -> pd.DataFrame:
    return energy_data_to_dataframe(energy_data)



import pytest

from CleanEmonCore.models import EnergyData
from CleanEmonBackend.lib.DBConnector import fetch_data
from CleanEmonBackend.lib.DBConnector import send_data
from CleanEmonBackend.lib.DBConnector import adapter

DUMMY_DATE = "2000-01-01"


@pytest.fixture
def energy_data():
    return EnergyData(DUMMY_DATE, [
        {"timestamp": 1, "power": 1, "temp": 1},
        {"timestamp": 2, "power": 2, "temp": 2},
        {"timestamp": 3, "power": 3, "temp": 3}
    ])


@pytest.mark.projectwise
def test_fetch_data():
    data = fetch_data("2022-05-01")
    assert data
    assert type(data) is EnergyData


@pytest.mark.projectwise
def test_clean_integration(energy_data):
    assert send_data(DUMMY_DATE, energy_data)
    assert energy_data == fetch_data(DUMMY_DATE)
    doc = adapter.get_document_id_for_date(DUMMY_DATE)
    assert doc
    assert adapter.delete_document(doc)

    assert not adapter.get_document_id_for_date(DUMMY_DATE)  # There should not be any changes in adapter




from CleanEmonBackend import NILM_INFERENCE_APIS_DIR
from CleanEmonBackend.lib.black_sorcery import nilm_path_fix


def test_nilm_path_fix():
    import os
    import sys

    with nilm_path_fix():
        assert os.getcwd() == NILM_INFERENCE_APIS_DIR
        assert NILM_INFERENCE_APIS_DIR in sys.path



import os

from CleanEmonCore.dotfiles import get_dotfile
from CleanEmonCore.dotfiles import DOT_DIR_PATH

DATA_DIR = os.path.join(DOT_DIR_PATH, "data")
RES_DIR = os.path.join(DOT_DIR_PATH, "res")

# Ensure that the data dir exists
if not os.path.isdir(DATA_DIR):
    os.mkdir(DATA_DIR)

CACHE_DIR = os.path.join(DATA_DIR, "cache")
PLOT_DIR = os.path.join(DATA_DIR, "plots")

# --- NILM-Inference-APIs ---
_NILM_CONFIG = "NILM-Inference-APIs.path"
NILM_CONFIG = get_dotfile(_NILM_CONFIG)
with open(NILM_CONFIG, "r") as f_in:
    NILM_INFERENCE_APIS_DIR = f_in.read().strip()

NILM_INPUT_DIR = os.path.join(NILM_INFERENCE_APIS_DIR, "input", "data")
if not os.path.exists(NILM_INPUT_DIR):
    os.makedirs(NILM_INPUT_DIR, exist_ok=True)

NILM_INPUT_FILE_PATH = os.path.join(NILM_INPUT_DIR, "data.csv")



import pytest
import json
import os


from CleanEmonCore.models import EnergyData
from CleanEmonBackend.lib.plots import plot_data


@pytest.fixture
def clean_data_df():
    directory = os.path.dirname(__file__)
    file = os.path.join(directory, "2022-05-15.json")

    with open(file, "r") as f_in:
        j = json.load(f_in)

    return EnergyData(date=j["date"], energy_data=j["energy_data"])


def test_plot_data(clean_data_df):
    plot_data(clean_data_df, columns=["timestamp", "power", "pred_fridge"])



import argparse

parser = argparse.ArgumentParser(prog="CleanEmonBackend", description="The CLI for CleanEmon-Backend")
subparsers = parser.add_subparsers(help="commands")

# Service
service_parser = subparsers.add_parser("service", help="Run a service")
service_parser.add_argument("service_name", action="store", choices=["api", "disaggregate"])

# Script
script_parser = subparsers.add_parser("script", help="Run a script")
script_parser.add_argument("script_name", action="store", choices=["disaggregate"])
script_parser.add_argument("dates", nargs="*",
                           help="list of dates (YYYY-MM-DD) to be used in `disaggregate` or `reset`")
script_parser.add_argument("--no-safe", action="store_false", default=False,
                           help="prompt before proceeding with critical actions")

# Setup
setup_parser = subparsers.add_parser("setup", help="Setup the backend system")
setup_parser.add_argument("setup_name", action="store", choices=["nilm"])
args = parser.parse_args()

if "service_name" in args:
    if args.service_name == "api":
        from .API.service import run
        run()
    elif args.service_name == "disaggregate":
        from .Disaggregator.service import run
        run()

elif "script_name" in args:
    if args.script_name == "disaggregate":
        from CleanEmonBackend.scripts.disaggregate import disaggregate
        if args.dates:
            disaggregate(*args.dates, no_prompt=args.no_safe)
        else:
            print("You should provide at least one date")

elif "setup_name" in args:
    if args.setup_name == "nilm":
        from . import NILM_CONFIG
        from CleanEmonBackend.scripts.setup import generate_nilm_inference_apis_config

        generate_nilm_inference_apis_config(NILM_CONFIG)



from datetime import datetime


def is_valid_date(date: str):
    try:
        datetime.strptime(date, "%Y-%m-%d")
        return True

    except ValueError:
        return False


def is_valid_date_range(from_date, to_date):
    try:
        dt_from = datetime.strptime(from_date, "%Y-%m-%d")
        dt_to = datetime.strptime(to_date, "%Y-%m-%d")
    except ValueError:
        return False

    return dt_to >= dt_from



import os
from typing import List, Tuple

import numpy as np
import pandas as pd

from .. import NILM_INPUT_FILE_PATH
from .. import NILM_INPUT_DIR
from ..lib.black_sorcery import nilm_path_fix


def _set_inference_input(df: pd.DataFrame) -> bool:
    df = df.copy()

    # Reformat as expected by NILM-Inference-APIs like "2022-05-16 00:00:17+01:00"
    df["Time"] = pd.to_datetime(df["Time"], unit='s')
    if not os.path.exists(NILM_INPUT_DIR):
        os.makedirs(NILM_INPUT_DIR, exist_ok=True)

    df.to_csv(NILM_INPUT_FILE_PATH, index=False)
    return True


def _disaggregate_to_files() -> List[Tuple[str, str]]:
    with nilm_path_fix():

        # ------------------------------------------------------#
        # Extremely ~spooky~ and error prune piece of code      #
        # Just for the history, this was deprecated since the   #
        # time of writing.                                      #
        #                                                       #
        # Uncle Bob, if you are reading, please forgive me :(   #
        # ------------------------------------------------------#
        from lab.nilm_trainer import nilm_inference
        from constants.enumerates import ElectricalAppliances

        # Consider all known devices
        devices = list(ElectricalAppliances)
        devices = [dev.value for dev in devices]

        return nilm_inference(devices=devices, sample_period=5, inference_cpu=True)


def disaggregate(df: pd.DataFrame, timestamp_label: str = "timestamp", target_label: str = "power") -> pd.DataFrame:
    df = df.copy()

    # Filter unneeded columns
    df_filtered = df.loc[:, [timestamp_label, target_label]]
    df_filtered[target_label] = df_filtered[target_label].fillna(df_filtered[target_label].mean().round())
    df_filtered = df_filtered.rename(columns={timestamp_label: "Time", target_label: "mains"})

    # Write dataframe to input file
    _set_inference_input(df_filtered)

    # Inference
    devices_files = _disaggregate_to_files()

    # Read data back into memory
    for device, file in devices_files:
        temp_df = pd.read_csv(file)
        first_n_missing = df.shape[0] - temp_df.shape[0]

        col_name = device.lower().replace(" ", "_")
        col_name = f"pred_{col_name}"

        preds = pd.concat([pd.Series(np.zeros(first_n_missing)), temp_df["preds"]])
        df[col_name] = preds.values
    # Clear rows that originally had NaN as target value, but keep timestamps
    df.loc[df[target_label].isna(), df.columns != timestamp_label] = np.NaN

    return df



class BadDateError(ValueError):
    def __init__(self, bad_date: str):
        self.bad_date = bad_date


class BadDateRangeError(ValueError):
    def __init__(self, bad_from_date: str, bad_to_date):
        self.bad_from_date = bad_from_date
        self.bad_to_date = bad_to_date



"""This module contains a set of utilities used to transform and prepare data for torch-nilm inference"""

import os
import json

from CleanEmonCore import CONFIG_FILE
from CleanEmonCore.models import EnergyData
from CleanEmonCore.CouchDBAdapter import CouchDBAdapter

from .. import CACHE_DIR

adapter = CouchDBAdapter(CONFIG_FILE)


def fetch_data(date_id: str, *, from_cache=False) -> EnergyData:

    if not os.path.exists(CACHE_DIR):
        os.mkdir(CACHE_DIR)

    name = date_id

    cache_path = os.path.join(CACHE_DIR, name)

    fetch_ok = False

    energy_data = EnergyData()

    if from_cache:
        try:
            with open(cache_path, "r") as fin:
                raw_data = json.load(fin)
                energy_data = EnergyData(raw_data["date"], raw_data["energy_data"])

            print("Fetched data from cache")
            fetch_ok = True
        except OSError:
            print("No cached data!")

    if not from_cache or not fetch_ok:
        energy_data = adapter.fetch_energy_data_by_date(date_id)

        # Cache data for future use
        with open(cache_path, "w") as fout:
            json.dump(energy_data.as_json(string=False), fout)

    return energy_data


def send_data(date_id: str, data: EnergyData):
    return adapter.update_energy_data_by_date(date_id, data)



import os


def generate_nilm_inference_apis_config(config_file):
    print("Configuration for NILM-Inference-APIs")
    print("-------------------------------------")
    nilm_path = input("Absolute path of NILM-Inference-APIs (the parent of its .git directory): ")

    if not os.path.isdir(nilm_path):
        print("Bad path. This will cause troubles in future. Please re-run")
        exit(1)
    else:
        with open(config_file, "w") as f_out:
            f_out.write(nilm_path)
        print(f"Config file was generated successfully at {config_file}")



from datetime import datetime
import json

import pandas as pd

from CleanEmonCore.models import EnergyData

INTERVAL = 5
INTERVAL_STR = f"{INTERVAL}S"
PERIODS = 60*60*24/INTERVAL


def reformat_timestamp(stamp: float) -> str:
    dt = datetime.fromtimestamp(stamp)
    return dt.strftime("%Y-%m-%d %H:%M:%S+01:00")


def revert_to_timestamp(f_datetime: str) -> str:
    dt = datetime.strptime(f_datetime, "%Y-%m-%d %H:%M:%S+01:00")
    return str(dt.timestamp())


def quantize_by_time(df: pd.DataFrame) -> pd.DataFrame:
    """Accepts a dataframe indexed by non-continuous datetime objects and returns a fully continuous dataframe.

    _Continuous_ refers to the series that has no gaps. Not to be confused with _continuous/discrete_ terms.

    Example:
    Say that INTERVAL=5. That means that every 5 seconds a new sample enters your dataframe. Ideally, by the end of the
    day you would have a dataframe with 17280 sample. This would look like so:
    -------Dataframe1------------
    00:00:00    values_0
    00:00:05    values_1
    00:00:10    values_2
    ...         ...
    23:59:45    values_17277
    23:59:50    values_17278
    23:59:50    values_17279
    -----------------------------

    Note that the values range from 0 up to 17279 (17280 in total). This is a fully continuous day, because no
    timeslots were missed.

    However, it is quite possible that your sampling system is not precise enough and some samples would fall out of
    that quantized pattern. The following dataframe is more likely to occur:
    -------Dataframe2------------
    00:00:03    values_0
    00:00:08    values_1
    00:00:12    values_2
    ...         ...
    23:59:45    values_17201
    23:59:49    values_17202
    23:59:57    values_17203
    -----------------------------

    Note that both indexes are not quantized (they don't repeat each 5 seconds) and the total number of samples is not
    17280, as some of them are never captured.

    This function would map Dataframe2 to Dataframe1. The workflow consists of those steps:
        1. Round each timestamp to its closest quantum
        2. If more than one original samples map to the same quantum, keep only one of them
        3. Quantum with no value still get to stay in their position even they may be empty (no sample mapped to them)
    """
    df.copy()
    # Copy some datetime details from the original dataframe
    today = df.index[0]
    tz = df.index[0].tz

    # Generate a new index consisting of continuous values
    continuous_index = pd.date_range(today, periods=PERIODS, freq=INTERVAL_STR, tz=tz)

    # Create the new dataframe with the continuous index
    new_df = pd.DataFrame(index=continuous_index, columns=df.columns)

    # Specify which columns can be filled using the old values
    intersection = new_df.index.intersection(df.index)

    # Fill the specified time slots with corresponding data from the old dataframe
    new_df.loc[intersection] = df.loc[intersection].values

    return new_df


def energy_data_to_dataframe(data: EnergyData, timestamp_label: str = "timestamp") -> pd.DataFrame:
    # Convert EnergyData to Dataframe
    df = pd.DataFrame(data.energy_data)

    # Keep original timestamp column as "original_timestamp"
    df[f"original_{timestamp_label}"] = df[timestamp_label].copy()

    df[timestamp_label] = pd.to_datetime(df[timestamp_label], unit='s')

    # # Reformat as expected by NILM-Inference-APIs like "2022-05-16 00:00:17+01:00"
    # df[timestamp_label] = df[timestamp_label].map(reformat_timestamp)

    # Index dataframe by time
    df[timestamp_label] = pd.DatetimeIndex(df[timestamp_label])
    df = df.set_index(timestamp_label)

    # Quantize time. Every timestamp should be mapped into fixed intervals.
    df.index = df.index.map(lambda date: date.round(INTERVAL_STR))

    # If more than one timestamp map into the same time quantum, arbitrarily remove all duplicates but the first
    df = df[~df.index.duplicated(keep="first")]

    # Generate a clean dataframe. All potentially empty time slots are filled
    df = quantize_by_time(df)

    # Reset index as expected by NILM-Inference-APIs
    df = df.reset_index()
    df = df.rename(columns={"index": timestamp_label})

    # Get back the python timestamp format
    df[timestamp_label] = df[timestamp_label].apply(lambda x: x.timestamp())

    # Convert to string
    df[timestamp_label] = df[timestamp_label].astype(str)

    return df


def dataframe_to_energy_data(df: pd.DataFrame, timestamp_label: str = "timestamp") -> EnergyData:

    # Ensure that timestamps are returned as floats, not strings
    df[timestamp_label] = df[timestamp_label].astype(float)
    first_timestamp = pd.to_datetime(df[timestamp_label], unit="s").iloc[0]
    date = first_timestamp.date()
    date = str(date)
    energy_data = json.loads(df.to_json(orient="records"))

    return EnergyData(date=date, energy_data=energy_data)



import os
from datetime import datetime

from typing import List

from matplotlib import pyplot as plt
import pandas as pd
import numpy as np

from CleanEmonCore.models import EnergyData

from .. import PLOT_DIR

if not os.path.exists(PLOT_DIR):
    os.makedirs(PLOT_DIR, exist_ok=True)


def timestamp_to_label(stamp):
    """Converts `stamp` into a datetime object and returns its reformatted string representation"""

    dt = datetime.fromtimestamp(stamp)

    return dt.strftime("%H:%M:%S")


def plot_data(energy_data: EnergyData, *, columns: List[str] = None, name="plot"):
    """Visualization the given dataframe"""

    df = pd.DataFrame(energy_data.energy_data)

    if not columns:
        columns = []
    else:
        columns = [col.lower() for col in columns]

    timestamp_label = "timestamp"

    if timestamp_label in df:
        time = df[timestamp_label]
        df = df.drop([timestamp_label], axis=1)
    else:
        time = df.index

    for col, data in df.iteritems():
        # Skip any timestamp-like label
        if "timestamp" in str(col).lower():
            continue
        # Filter only selected columns
        if (columns and str(col).lower() in columns) or (not columns):
            mask = np.isfinite(data)
            plt.plot(time[mask], data[mask], label=col)
    skip = len(time) // 15

    plt.subplots_adjust(bottom=0.25)
    plt.xticks(time[0:-1:skip], time[0:-1:skip].map(timestamp_to_label), rotation=90, fontsize=8)
    plt.legend()
    plt.title(energy_data.date)
    plt.xlabel("Time")
    fout_name = os.path.join(PLOT_DIR, f"{name}.png")
    plt.savefig(fout_name, dpi=1000)
    plt.cla()
    plt.clf()
    return fout_name




"""The definition of Black Sorcery
This module provides the stupidest fixes that are ment to be used *if and only if* there are no other sound workarounds.

Deep down in my heart, I kinna like them tho,
George Vasiliadis
"""

from contextlib import contextmanager

from CleanEmonBackend import NILM_INFERENCE_APIS_DIR


@contextmanager
def nilm_path_fix(nilm_path=NILM_INFERENCE_APIS_DIR):
    """NILM-Inference-APIs is ment to be run as a standalone set of scripts and not as a module (e.g. absence of
    relative imports). In order to integrate it in other packages and/or modules, one must always change manually the
    execution directory and the system path back and forth. This context manager ensures that this operation is done
    consistently.

    Example of use:
    >>> with nilm_path_fix(NILM_INFERENCE_APIS_PATH):
    >>>     # We are virtually executing code from within NILM-Inference-APIs directory
    >>>     function_that_does_nilm_related_stuff()
    >>> # We are back to normal namespace and paths-state
    """

    import os
    import sys
    _original_cwd = os.getcwd()
    _original_path = sys.path.copy()

    os.chdir(nilm_path)
    sys.path.append(nilm_path)
    yield

    os.chdir(_original_cwd)
    sys.path = _original_path
